/*
 * olednd.cpp --
 *
 *	Initialize an interpreter with OLE drag and drop extensions.
 *	This mostly just makes a bunch of calls to Tcl_CreateCommand to
 *	create tcl commands for the various modules.  This file is mainly
 *	written in C with a few C++ uses to interface with the completely
 *	C++ OLE interface objects.
 *
 * Copyright (c) 1995 LAS Incorporated.
 * All rights reserved.
 *
 * Copyright (c) 1995-1998 University of California, Berkeley.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * IN NO EVENT SHALL LAS INCORPORATED BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * LAS INCORPORATED SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND LAS INCORPORATED HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 * 
 * Notes: XXX
 *	1. Lots of functions now take event arguments in the command
 *	   handling code.
 *	2. New events: <Drop> <Drag> <Enter> <Leave>?
 */

#include <windows.h>
#include <ole2.h>

extern "C" {
#include "tk.h"
#include "tkWin.h"
}
#include "tobject.h"

static Tcl_HashTable targetTable;
static Tcl_HashTable sourceTable;
static int initialized = 0;

typedef struct _DndType {
    int priority;		/* For target types, check priorities */
    CLIPFORMAT type;		/* Clipboard format */
    char *typeStr;		/* Name of type */
    char *eventStr;		/* Name of event */
    char *script;		/* Script to run */
    struct _DndType *next;	/* Next one in list */
} DndType;

typedef struct _DndInfo {
    Tcl_Interp *interp;		/* The associated interp */
    Tk_Window topwin;		/* The main Tk window */
    Tk_Window tkwin;		/* The associated Tk window */
    DndType head;		/* Head of the list of supported types */
    LPDROPTARGET IDropTarget;	/* The OLE IDropTarget object */
    DndType *cbData;
    PTCallback callbacks;
    Tcl_HashEntry *hashEntry;	/* Hash table entry */
} DndInfo;

/*
 * Forward declarations for procedures defined later in this file:
 */
static int	AddHandler(Tcl_Interp *interp, Tk_Window topwin,
		    Tcl_HashTable *table, char *windowPath, char *typeStr,
		    char *eventStr, char *script,
		    int isDropTarget, int priority);
static int	DelHandler(DndInfo *infoPtr, char *typeStr, char *eventStr);
static int	DelHandlerByName(Tcl_Interp *interp, Tk_Window topwin,
		    Tcl_HashTable *table, char *windowPath,
		    char *typeStr, char *eventStr);
static void	DestroyEventProc(ClientData clientData, XEvent *eventPtr);
static int	DndDrag(Tcl_Interp *interp, char *windowPath);
static int	GetCurrentScript(Tcl_Interp *interp, Tk_Window topwin,
		    Tcl_HashTable *table, char *windowPath,
		    char *typeStr, char *eventStr);
static int	GetCurrentTypes(Tcl_Interp *interp, Tk_Window topwin,
		    Tcl_HashTable *table, char *windowPath);

/*
 *----------------------------------------------------------------------
 *
 * DestroyEventProc --
 *
 *	This routine gets called on events that match the StructureNotify
 *	mask.  We only care about DestroyNotify events.
 *
 * Results:
 *	None
 *
 * Side Effects:
 *	Removes the window from being either a source or a target.
 *
 *----------------------------------------------------------------------
 */
static void
DestroyEventProc(ClientData clientData, XEvent *eventPtr)
{
    DndInfo *infoPtr = (DndInfo *) clientData;

    if (eventPtr->type == DestroyNotify) {
	DelHandler(infoPtr, NULL, NULL);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * DelHandler --
 *
 *	Delete a handler.  Assumes everything has already been looked
 *	up elsewhere
 *
 * Results:
 *	A standard TCL results
 *
 *----------------------------------------------------------------------
 */

static int
DelHandler(DndInfo *infoPtr, char *typeStr, char *eventStr)
{
    DndType *head, *prev, *curr, *next;
    int match;
    int eventLen;

    if (eventStr) {
	eventLen = lstrlen(eventStr);
    } else {
	eventLen = 0;
    }
	
    head = prev = &infoPtr->head;
    for (curr = head->next; curr != NULL; curr = next) {
	next = curr->next;
	match = 0;
	if ((typeStr == NULL) || (lstrcmp(curr->typeStr, typeStr) == 0)) {
	    if (eventLen) {
		if (lstrcmp(curr->eventStr, eventStr) == 0) {
		    match = 1;
		}
	    } else {
		match = 1;
	    }
	    if (match) {
		ckfree(curr->typeStr);
		ckfree(curr->script);
		prev->next = next;
	    }
	} else {
	    prev = curr;
	}
    }

    if (head->next == NULL) {
	if (infoPtr->IDropTarget) {
	    RevokeDragDrop(Tk_GetHWND(Tk_WindowId(infoPtr->tkwin)));
	    Tk_DeleteEventHandler(infoPtr->tkwin, StructureNotifyMask,
				  DestroyEventProc, (ClientData) infoPtr);
	    infoPtr->IDropTarget->Release();
	    infoPtr->callbacks->Release();
	}

	ckfree((char *) infoPtr);
	Tcl_DeleteHashEntry(infoPtr->hashEntry);
    }

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * DelHandlerByName --
 *
 *	Remove a drag&drop handler window, freeing up any memory and objects
 *	that need to be cleaned up.
 *
 * Arguments:
 *	interp:		Tcl interpreter
 *	topwin:		The main window
 *	table:		Hash table we are removing the window/type from
 *	windowPath:	String pathname of window
 *	typeStr:	The data type to remove.
 *	eventStr:	Event name to remove
 *
 * Results:
 *	A standard Tcl result
 *
 * Notes:
 *	The table that is passed in must have been initialized.
 *	If type == NULL, then everything is cleaned up.
 *
 *----------------------------------------------------------------------
 */
static int
DelHandlerByName(Tcl_Interp *interp, Tk_Window topwin, Tcl_HashTable *table,
    char *windowPath, char *typeStr, char *eventStr)
{
    Tcl_HashEntry *hPtr;
    Tk_Window tkwin;

    tkwin = Tk_NameToWindow(interp, windowPath, topwin);
    if (tkwin == NULL) {
	return TCL_ERROR;
    }

    hPtr = Tcl_FindHashEntry(table, windowPath);
    if (hPtr == NULL) {
	return TCL_OK;
    }

    return DelHandler((DndInfo *) Tcl_GetHashValue(hPtr), typeStr, eventStr);
}

/*
 *----------------------------------------------------------------------
 *
 * ExpandPercents --
 *
 *	Given a command and an event, produce a new command
 *	by replacing % constructs in the original command
 *	with information from the drop event
 *
 * Results:
 *	The new expanded command is appended to the dynamic string
 *	given by dsPtr.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */
static void
ExpandPercents(DndInfo *infoPtr, DndType *typePtr,
	       char *before, Tcl_DString *dsPtr, LONG x, LONG y)
{
    int spaceNeeded, cvtFlags;	/* Used to substitute string as proper Tcl
				 * list element. */
    int number, length;
#define NUM_SIZE 40
    register char *string;
    char numStorage[NUM_SIZE+1];
    POINT sPt;
    int converted = 0;

    while (1) {
	/*
	 * Find everything up to the next % character and append it
	 * to the result string.
	 */

	for (string = before; (*string != 0) && (*string != '%'); string++) {
	    /* Empty loop body. */
	}
	if (string != before) {
	    Tcl_DStringAppend(dsPtr, before, string-before);
	    before = string;
	}
	if (*before == 0) {
	    break;
	}

	/*
	 * There's a percent sequence here.  Process it.
	 */

	number = 0;
	string = "??";
	switch (before[1]) {
	  case 'T':
	    string = typePtr->typeStr;
	    goto doString;
	  case 'W':
	    string = Tk_PathName(infoPtr->tkwin);
	    goto doString;
	  case 'X':
	    number = x;
	    goto doNumber;
	  case 'Y':
	    number = y;
	    goto doNumber;
	  case 'x':
	    if (! converted) {
		sPt.x = x;
		sPt.y = y;
		ScreenToClient(Tk_GetHWND(Tk_WindowId(infoPtr->tkwin)), &sPt);
	    }
	    number = sPt.x;
	    goto doNumber;

	  case 'y':
	    if (! converted) {
		sPt.x = x;
		sPt.y = y;
		ScreenToClient(Tk_GetHWND(Tk_WindowId(infoPtr->tkwin)), &sPt);
	    }
	    number = sPt.y;
	    goto doNumber;

	  default:
	    numStorage[0] = before[1];
	    numStorage[1] = '\0';
	    string = numStorage;
	    goto doString;
	}
	doNumber:
	wsprintf(numStorage, "%d", number);
	string = numStorage;

	doString:
	spaceNeeded = Tcl_ScanElement(string, &cvtFlags);
	length = Tcl_DStringLength(dsPtr);
	Tcl_DStringSetLength(dsPtr, length + spaceNeeded);
	spaceNeeded = Tcl_ConvertElement(string,
		Tcl_DStringValue(dsPtr) + length,
		cvtFlags | TCL_DONT_USE_BRACES);
	Tcl_DStringSetLength(dsPtr, length + spaceNeeded);
	before += 2;
    }
}	    


/*
 *----------------------------------------------------------------------
 *
 * TCallback::Reset --
 *
 *	This routine is called from TDropTarget.
 *
 * Results:
 *	S_OK
 *
 *----------------------------------------------------------------------
 */
HRESULT
TCallback::Reset(LPVOID clientData)
{
    DndInfo *infoPtr = (DndInfo *) clientData;

    infoPtr->cbData = infoPtr->head.next;

    return S_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * TCallback::Next --
 *
 *	This routine is called from TDropTarget.
 *
 * Results:
 *	S_OK
 *
 *----------------------------------------------------------------------
 */
HRESULT
TCallback::Next(LPVOID clientData, CLIPFORMAT *pcfFormatOut)
{
    DndInfo *infoPtr = (DndInfo *) clientData;

    if (infoPtr->cbData) {
	*pcfFormatOut = infoPtr->cbData->type;
	infoPtr->cbData = infoPtr->cbData->next;
	return S_OK;
    }
    return ResultFromScode(S_FALSE);
}

/*
 *----------------------------------------------------------------------
 *
 * TCallback::GetData --
 *
 *	This routine is called from TDropTarget.
 *	If the format is CF_TEXT, then we need to create a version of the
 *	string with \r\n as the line return value instead of just \n.
 *
 * Results:
 *	S_OK
 *
 *----------------------------------------------------------------------
 */
HRESULT
TCallback::GetData(LPVOID clientData, UINT cfFormat, char **ppStrOut)
{
    DndInfo *infoPtr = (DndInfo *) clientData;
    int ret;
    DndType *curr;
    Tcl_DString dString;
    int len;
    char *s, *ip, *op;

    for (curr = infoPtr->head.next; curr != NULL; curr = curr->next) {
	if (curr->type == cfFormat) {
	    Tcl_DStringInit(&dString);
	    ExpandPercents(infoPtr, curr, curr->script, &dString, 0, 0);
	    ret = Tcl_GlobalEval(infoPtr->interp, dString.string);
	    Tcl_DStringFree(&dString);
	    if (ret == TCL_ERROR) {
		Tk_BackgroundError(infoPtr->interp);
	    } else if (ret == TCL_BREAK) {
		return S_FALSE;
	    }
	    if (cfFormat == CF_TEXT) {
		len = lstrlen(infoPtr->interp->result);
		s = (char *) ckalloc(len * 2 + 1);
		for (ip = infoPtr->interp->result, op = s; *ip != 0; ip++) {
		    if (*ip == '\n') {
			*op++ = '\r';
		    }
		    *op++ = *ip;
		}
		*op = 0;
		Tcl_SetResult(infoPtr->interp, s, TCL_DYNAMIC);
	    }
	    *ppStrOut = infoPtr->interp->result;
	    return S_OK;
	}
    }
    return ResultFromScode(S_FALSE);
}

/*
 *----------------------------------------------------------------------
 *
 * TCallback::SetData --
 *
 *	This routine is called from TDropTarget.
 *
 * Results:
 *	S_OK
 *
 *----------------------------------------------------------------------
 */

HRESULT
TCallback::SetData(LPVOID clientData, UINT cfFormat, LONG x, LONG y,
    char *str, char *eventStr)
{
    DndInfo *infoPtr = (DndInfo *) clientData;
    int ret;
    DndType *curr;
    Tcl_DString dString;
    char *s, *ip, *op;
    int len;
    int knownType = 0;

    s = NULL;
    for (curr = infoPtr->head.next; curr != NULL; curr = curr->next) {
	if (curr->type != cfFormat) {
	    continue;
	}
	knownType = 1;
	if (lstrcmp(curr->eventStr, eventStr) != 0) {
	    continue;
	}

	Tcl_DStringInit(&dString);
	ExpandPercents(infoPtr, curr, curr->script, &dString, x, y);

	if (str) {
	    if (cfFormat == CF_TEXT) {
		len = lstrlen(str) + 1;
		s = (char *) ckalloc(len);
		memcpy(s, str, len);

		for (ip = str, op = s; *ip != 0; ) {
		    if (*ip == '\r') {
			if (ip[1] != '\n') {
			    *op++ = '\n';
			}
			ip++;
		    } else {
			*op++ = *ip++;
		    }
		}
		*op = 0;
		Tcl_DStringAppendElement(&dString, s);
		ckfree(s);
	    } else {
		Tcl_DStringAppendElement(&dString, str);
	    }
	}

	ret = Tcl_GlobalEval(infoPtr->interp, dString.string);
	Tcl_DStringFree(&dString);
	if (ret == TCL_ERROR) {
	    Tk_BackgroundError(infoPtr->interp);
	}
	return S_OK;
    }
    if (s != NULL) {
	ckfree(s);
    }
    if (! knownType) {
	return ResultFromScode(S_FALSE);
    } else {
	return S_OK;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * AddHandler --
 *
 *	Add a drag&drop handler to a window, freeing up any memory and
 *	objects that need to be cleaned up.
 *
 * Arguments:
 *	interp:		Tcl interpreter
 *	topwin:		The main window
 *	table:		Hash table we are adding the window/type to
 *	windowPath:	String pathname of window
 *	typeStr:	The data type to add.
 *      script:	        Script to run
 *	isDropTarget:	Is this target a drop target?
 *	priority:	Priority of this type
 *
 * Results:
 *	A standard Tcl result
 *
 * Notes:
 *	The table that is passed in must have been initialized.
 *	Inserts a new node in the typelist for the window with the
 *	highest priorities listed first.
 *
 *----------------------------------------------------------------------
 */
static int
AddHandler(Tcl_Interp *interp, Tk_Window topwin, Tcl_HashTable *table,
	   char *windowPath, char *typeStr, char *eventStr, char *script,
	   int isDropTarget, int priority)
{
    Tcl_HashEntry *hPtr;
    DndInfo *infoPtr;
    int created;
    Tk_Window tkwin;
    DndType *head, *prev, *curr, *tnew;
    int len;
    UINT type;
    HRESULT hret;

    tkwin = Tk_NameToWindow(interp, windowPath, topwin);
    if (tkwin == NULL) {
	return TCL_ERROR;
    }
    Tk_MakeWindowExist(tkwin);

    if (lstrcmp(typeStr, "CF_TEXT") == 0) {
	type = CF_TEXT;
    } else if (lstrcmp(typeStr, "CF_UNICODETEXT") == 0) {
	type = CF_UNICODETEXT;
    } else {
	type = RegisterClipboardFormat(typeStr);
    }
    if (type == 0) {
	Tcl_AppendResult(interp,
			 "unable to register clipboard format ",
			 typeStr, (char *) NULL);
	return TCL_ERROR;
    }

    if (! isDropTarget) {
	priority = 0;
    }

    tnew = (DndType *) ckalloc(sizeof(DndType));
    tnew->priority = priority;
    tnew->type = type;
    len = lstrlen(typeStr) + 1;
    tnew->typeStr = (char *) ckalloc(len);
    memcpy(tnew->typeStr, typeStr, len);
    len = lstrlen(eventStr) + 1;
    tnew->eventStr = (char *) ckalloc(len);
    memcpy(tnew->eventStr, eventStr, len);
    len = lstrlen(script) + 1;
    tnew->script = (char *) ckalloc(len);
    memcpy(tnew->script, script, len);
    tnew->next = NULL;

    hPtr = Tcl_CreateHashEntry(table, windowPath, &created);
    if (created) {
	infoPtr = (DndInfo *) ckalloc(sizeof(DndInfo));
	infoPtr->head.next = NULL;
	infoPtr->interp = interp;
	infoPtr->tkwin = tkwin;
	infoPtr->hashEntry = hPtr;
	infoPtr->callbacks = new TCallback();
	if (isDropTarget) {
	    infoPtr->IDropTarget = new TDropTarget(infoPtr->callbacks,
						   (LPVOID) infoPtr);
	    infoPtr->IDropTarget->AddRef();
	    hret = RegisterDragDrop(Tk_GetHWND(Tk_WindowId(tkwin)), infoPtr->IDropTarget);
	    if (hret != S_OK) {
		infoPtr->IDropTarget->Release();
		ckfree((char *) infoPtr);
		Tcl_AppendResult(interp, "unable to register \"",
				 windowPath,
				 "\" as a droppable window", (char *) NULL);
		Tcl_DeleteHashEntry(hPtr);
		return TCL_ERROR;
	    }
	} else {
	    infoPtr->IDropTarget = NULL;
	}
	Tk_CreateEventHandler(tkwin, StructureNotifyMask,
			      DestroyEventProc, (ClientData) infoPtr);
	infoPtr->callbacks->AddRef();

	Tcl_SetHashValue(hPtr, infoPtr);
	infoPtr->head.next = tnew;

    } else {
	infoPtr = (DndInfo *) Tcl_GetHashValue(hPtr);
	head = prev = &infoPtr->head;
	for (curr = head; curr != NULL; prev = curr, curr = curr->next) {
	    if (curr->priority > priority) {
		break;
	    }
	}
	tnew->next = prev->next;
	prev->next = tnew;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 * GetCurrentTypes --
 *
 *	Append a list of the current types that are handled by the given
 *	to the result.
 *
 * Results:
 *	A standard TCL result.
 *
 *----------------------------------------------------------------------
 */
static int
GetCurrentTypes(Tcl_Interp *interp, Tk_Window topwin, Tcl_HashTable *table,
		char *windowPath)
{
    Tcl_HashEntry *hPtr;
    DndInfo *infoPtr;
    Tk_Window tkwin;
    DndType *curr;

    tkwin = Tk_NameToWindow(interp, windowPath, topwin);
    if (tkwin == NULL) {
	return TCL_ERROR;
    }

    hPtr = Tcl_FindHashEntry(table, windowPath);
    if (hPtr == NULL) {
	return TCL_OK;
    }

    infoPtr = (DndInfo *) Tcl_GetHashValue(hPtr);
    for (curr = infoPtr->head.next; curr != NULL; curr = curr->next) {
	Tcl_AppendElement(interp, curr->typeStr);
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 * GetCurrentScript --
 *
 *	Append the current script for the type handler of the given
 *	window.
 *
 * Results:
 *	A standard TCL result.
 *
 *----------------------------------------------------------------------
 */
static int
GetCurrentScript(Tcl_Interp *interp, Tk_Window topwin, Tcl_HashTable *table,
		 char *windowPath, char *typeStr, char *eventStr)
{
    Tcl_HashEntry *hPtr;
    DndInfo *infoPtr;
    Tk_Window tkwin;
    DndType *curr;

    tkwin = Tk_NameToWindow(interp, windowPath, topwin);
    if (tkwin == NULL) {
	return TCL_ERROR;
    }

    hPtr = Tcl_FindHashEntry(table, windowPath);
    if (hPtr == NULL) {
	return TCL_OK;
    }

    infoPtr = (DndInfo *) Tcl_GetHashValue(hPtr);
    for (curr = infoPtr->head.next; curr != NULL; curr = curr->next) {
	if (lstrcmp(curr->typeStr, typeStr) == 0 &&
	    lstrcmp(curr->eventStr, eventStr) == 0)
	{
	    Tcl_AppendElement(interp, curr->script);
	    return TCL_OK;
	}
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 * DndDrag --
 *	Begins an OLE drag and returns when the drop finishes or the
 *	operation is cancelled.
 *
 * Results:
 *	Standard TCL result
 *
 * Notes:
 *	windowPath is assumed to be a valid Tk window pathname at this
 *	point.
 *----------------------------------------------------------------------
 */
static int
DndDrag(Tcl_Interp *interp, char *windowPath, int button)
{
    LPDROPSOURCE pDropSource;
    PTDataObject pDataObject;
    PTCallback pCallbacks;
    DWORD keyState;
    Tcl_HashEntry *hPtr;
    DndInfo *infoPtr;
    DndType *curr;
    DWORD actualEffect;
    HRESULT ret;
    POINT currPt;
    HWND hwnd;
    DWORD mkButton;

    hPtr = Tcl_FindHashEntry(&sourceTable, windowPath);
    if (hPtr == NULL) {
	Tcl_AppendResult(interp, "unable to begin drag operation: ",
			 "no source types have been bound to ", windowPath,
			 (char *) NULL);
	return TCL_ERROR;
    }

    infoPtr = (DndInfo *) Tcl_GetHashValue(hPtr);

    switch (button) {
      case 1:
	mkButton = MK_LBUTTON;
	break;
      case 2:
	mkButton = MK_MBUTTON;
	break;
      case 3:
	mkButton = MK_RBUTTON;
	break;
      default:
	mkButton = 0;
    }

    keyState = 0;
    if (GetKeyState(VK_SHIFT)   & 0x8000) keyState |= MK_SHIFT;
    if (GetKeyState(VK_CONTROL) & 0x8000) keyState |= MK_CONTROL;
    if (GetKeyState(VK_MENU)    & 0x8000) keyState |= MK_ALT;
    if (GetKeyState(VK_LBUTTON) & 0x8000) keyState |= MK_LBUTTON;
    if (GetKeyState(VK_MBUTTON) & 0x8000) keyState |= MK_MBUTTON;
    if (GetKeyState(VK_RBUTTON) & 0x8000) keyState |= MK_RBUTTON;

    if (button && (!(keyState & mkButton))) {
	return TCL_OK;
    }
    pDropSource = new TDropSource(keyState, mkButton); 
    pDropSource->AddRef();
    if (pDropSource == NULL) {
	Tcl_AppendResult(interp, "unable to create OLE TDropSource",
			 (char *) NULL);
	return TCL_ERROR;
    }

    pCallbacks = new TCallback();
    pCallbacks->AddRef();
    pDataObject = new TDataObject(pCallbacks, (LPVOID) infoPtr);
    pDataObject->AddRef();
    if (pDataObject == NULL) {
	Tcl_AppendResult(interp, "unable to create OLE CDataObject",
			 (char *) NULL);
	return TCL_ERROR;
    }

    for (curr = infoPtr->head.next; curr != NULL; curr = curr->next) {
	pDataObject->AddDataType(curr->type);
    }

    ret = DoDragDrop(pDataObject, pDropSource, DROPEFFECT_COPY, &actualEffect);

    pDataObject->Release();
    pDropSource->Release();
    pCallbacks->Release();

    GetCursorPos(&currPt);
    hwnd = WindowFromPoint(currPt);
    if ((keyState & MK_LBUTTON) && (! (GetKeyState(VK_LBUTTON) & 0x8000))) {
	PostMessage(hwnd, WM_LBUTTONUP, 0, MAKELONG(currPt.x, currPt.y));
    }
    if ((keyState & MK_MBUTTON) && (! (GetKeyState(VK_MBUTTON) & 0x8000))) {
	PostMessage(hwnd, WM_MBUTTONUP, 0, MAKELONG(currPt.x, currPt.y));
    }
    if ((keyState & MK_RBUTTON) && (! (GetKeyState(VK_RBUTTON) & 0x8000))) {
	PostMessage(hwnd, WM_RBUTTONUP, 0, MAKELONG(currPt.x, currPt.y));
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 * Tole_DndCmd --
 *
 *	Drag and drop command.  This has several forms.
 *
 * Results:
 *	A standard TCL result.
 *----------------------------------------------------------------------
 */
int
Tole_DndCmd(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Tk_Window topwin = (Tk_Window) clientData;
    Tk_Window tkwin;
    int c, length;
    int button = 0;

    if (argc < 3) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
			 " option window ?arg ...?\"", (char *) NULL);
	return TCL_ERROR;
    }

    if (! initialized) {
	Tcl_InitHashTable(&targetTable, TCL_STRING_KEYS);
	Tcl_InitHashTable(&sourceTable, TCL_STRING_KEYS);
	initialized = 1;
    }

    tkwin = Tk_NameToWindow(interp, argv[2], topwin);
    if (tkwin == NULL) {
	return TCL_ERROR;
    }

    c = argv[1][0];
    length = lstrlen(argv[1]);

    if ((c == 'b') && (strncmp(argv[1], "bindsource", length) == 0)
	    && (length >= 5)) {
	if (argc == 3) {
	    return GetCurrentTypes(interp, topwin, &sourceTable, argv[2]);
	}
	if (argc == 4) {
	    return GetCurrentScript(interp, topwin, &sourceTable,
				    argv[2], argv[3], "");
	}
	if (argc == 5) {
	    if (lstrlen(argv[4]) > 0) {
		return AddHandler(interp, topwin, &sourceTable,
				  argv[2], argv[3], "", argv[4], 0, 0);
	    } else {
		return DelHandlerByName(interp, topwin, &sourceTable,
				  argv[2], argv[3], NULL);
	    }
	}
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " bindsource window",
			 " ?type ?script ?-priority number???\"",
			 (char *) NULL);
	return TCL_ERROR;
    } else if ((c == 'b') && (strncmp(argv[1], "bindtarget", length) == 0)
	    && (length >= 5)) {
	int priority;
	int arg;
	char *event;

	if (argc == 3) {
	    return GetCurrentTypes(interp, topwin, &targetTable, argv[2]);
	}

	event = "";
	arg = 4;
	if (argc >= 5 && argv[arg][0] == '<') {
	    event = argv[arg];
	    arg++;
	} else {
	    event = "<Drop>";
	}
	if (argc == arg) {
	    return GetCurrentScript(interp, topwin, &targetTable,
				    argv[2], argv[3], event);
	}

	arg++;
	if (argc == arg) {
	    if (lstrlen(argv[arg-1]) > 0) {
		return AddHandler(interp, topwin, &targetTable,
				  argv[2], argv[3], event, argv[arg-1], 1, 50);
	    } else {
		return DelHandlerByName(interp, topwin, &targetTable,
				  argv[2], argv[3], event);
	    }
	}
	if (argc != arg+2) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " bindtarget window",
			     " ?type ?<event>? ?script ?-priority number???\"",
			     (char *) NULL);
	    return TCL_ERROR;
	}
	if (lstrcmp(argv[arg], "-priority") != 0) {
	    Tcl_AppendResult(interp, "unknown option \"", argv[arg],
			     "\": must be -priority", (char *) NULL);
	    return TCL_ERROR;
	}
	arg++;
	if (Tcl_GetInt(interp, argv[arg], &priority) != TCL_OK) {
	    return TCL_ERROR;
	}
	if (priority < 1) {
	    priority = 1;
	} else if (priority > 100) {
	    priority = 100;
	}
	return AddHandler(interp, topwin, &targetTable, argv[2], argv[3],
			  event, argv[arg-2], 1, priority);
    } else if ((c == 'c') && (strncmp(argv[1], "clearsource", length) == 0)
	    && (length >= 6)) {
	return DelHandlerByName(interp, topwin, &sourceTable, argv[2], NULL, NULL);
    } else if ((c == 'c') && (strncmp(argv[1], "cleartarget", length) == 0)
	    && (length >= 6)) {
	return DelHandlerByName(interp, topwin, &targetTable, argv[2], NULL, NULL);
    } else if ((c == 'd') && (strncmp(argv[1], "drag", length) == 0)) {
	if (argc != 3 && argc != 5) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " drag window\"", (char *) NULL);
	    return TCL_ERROR;
	}
	if (argc == 5) {
	    if (lstrcmp(argv[3], "-button") != 0) {
		Tcl_AppendResult(interp, "unknown option \'", argv[3],
				 "\': must be -button",
				 (char *) NULL);
	    }
	    if (Tcl_GetInt(interp, argv[4], &button) != TCL_OK) {
		return TCL_ERROR;
	    }
	}
	return DndDrag(interp, argv[2], button);
    } else {
	Tcl_AppendResult(interp, "unknown or ambiguous option \"", argv[1],
		"\": must be bindsource, bindtarget, clearsource, or ",
		"cleartarget", (char *) NULL);
	return TCL_ERROR;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Tkolednd_Init --
 *
 *	Initialize the OLE drag and drop package as well as OLE2.
 *
 * Results:
 *	A standard Tcl result
 *
 * Side effects:
 *	OLE command are bound into interpreter.
 *
 *----------------------------------------------------------------------
 */
extern "C"
int
Tkolednd_Init(Tcl_Interp *interp)
{
    Tk_Window topwin;
    HRESULT hret;

    topwin = Tk_MainWindow(interp);
    if (topwin == NULL) {
	return TCL_ERROR;
    }

    hret = OleInitialize(NULL);

    /*
     * If OleInitialize returns S_FALSE, OLE has already been initialized
     */
    if (hret != S_OK && hret != S_FALSE) {
	Tcl_AppendResult(interp, "unable to initialize OLE2",
			 (char *) NULL);
	return TCL_ERROR;
    }
    Tcl_CreateCommand (interp, "olednd", Tole_DndCmd,
	   (ClientData) topwin, (void (*) _ANSI_ARGS_((ClientData))) NULL);

    return TCL_OK;
}

extern "C"
int
Tkolednd_SafeInit(Tcl_Interp *interp)
{
    return Tkolednd_Init(interp);
}
