/*
 *----------------------------------------------------------------------
 * Filename: tdataobject.cpp
 *
 *	This file implements the TDataObject.  Right now, it only defines
 *	the data transfer parts of the interface.
 *
 * Copyright (c) 1995 LAS Incorporated.
 * All rights reserved.
 *
 * Copyright (c) 1995 University of California, Berkeley.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * IN NO EVENT SHALL LAS INCORPORATED BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * LAS INCORPORATED SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND LAS INCORPORATED HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *----------------------------------------------------------------------
 */

#include <windows.h>
#include <string.h>
#include "tobject.h"

/*
 *----------------------------------------------------------------------
 * TDataObject::TDataObject
 *
 *	Constructor
 *----------------------------------------------------------------------
 */
TDataObject::TDataObject(PTCallback callbacks, LPVOID clientData)
{
    m_callbacks = callbacks;
    m_clientData = clientData;
    m_refCnt = 0;
    m_numTypes = 0;
    m_maxTypes = 0;
}

/*
 *----------------------------------------------------------------------
 * TDataObject::~TDataObject
 *
 *	Destructor
 *----------------------------------------------------------------------
 */
TDataObject::~TDataObject(void)
{
    if (m_numTypes > 0) {
	delete [] m_typeList;
    }
}

/*
 *----------------------------------------------------------------------
 * TDataObject::AddDataType
 *
 *	Add a data type to the list of types that are supported by
 *	this instance of the TDataObject.
 *
 * Results:
 *	0 if everything is OK.  Error if not.
 *
 * Notes:
 *	This routine does not check if an existing type matches the
 *	added type.
 *----------------------------------------------------------------------
 */
HRESULT
TDataObject::AddDataType(UINT clipFormat)
{
    FORMATETC *oldTypeList;
    UINT i;

    if (clipFormat == 0) {
	return 1;
    }
    if (m_maxTypes == m_numTypes) {
	m_maxTypes += 10;
	oldTypeList = m_typeList;
	m_typeList = new FORMATETC[m_maxTypes];
	if (m_typeList == NULL) {
	    return ERROR_NOT_ENOUGH_MEMORY;
	}
	if (m_numTypes > 0) {
	    for (i = 0; i < m_numTypes; i++) {
		m_typeList[i] = oldTypeList[i];
	    }
	    delete [] oldTypeList;
	}
    }

    m_typeList[m_numTypes].cfFormat = clipFormat;
    m_typeList[m_numTypes].ptd = NULL;
    m_typeList[m_numTypes].dwAspect = DVASPECT_CONTENT;
    m_typeList[m_numTypes].lindex = -1;
    m_typeList[m_numTypes].tymed = TYMED_HGLOBAL;

    m_numTypes++;
    return 0;
}

/*
 *----------------------------------------------------------------------
 * TDataObject::DelDataType
 *
 *	Remove a data type from the list of types that are supported by
 *	this instance of the TDataObject.
 *
 * Results:
 *	0 if the type was found in the list and was successfully removed.
 *	A nonzero value if the type was not found.
 *
 * Notes:
 *	If there is more than one of the same clipFormats in the list,
 *	only the first one is removed.
 *----------------------------------------------------------------------
 */
HRESULT
TDataObject::DelDataType(UINT clipFormat)
{
    UINT i, j;

    if (clipFormat == 0) {
	return 1;
    }
    for (i = 0; i < m_numTypes; i++) {
	if (m_typeList[i].cfFormat == clipFormat) {
	    for (j = i; j < m_numTypes; j++) {
		m_typeList[j-1] = m_typeList[j];
	    }
	    m_numTypes--;
	    return 0;
	}
    }
    return 1;
}

/*
 *----------------------------------------------------------------------
 * TDataObject::QueryInterface
 *
 *	Does interface negotiation.
 *
 * Arguments:
 *
 *	REFIID riid	A reference to the interface being queried.
 *      LPVOID *ppvObj	Out parameter to return a pointer to interface.
 *
 * Results:
 *	S_OK if interface is supported, E_NOINTERFACE if it is not
 *
 *----------------------------------------------------------------------
 */
STDMETHODIMP
TDataObject::QueryInterface(REFIID riid, LPVOID *ppvObj)
{
    if (IsEqualIID(riid, IID_IUnknown) || IsEqualIID(riid, IID_IDataObject)) {
	AddRef();
	*ppvObj = this;
	return S_OK;
    } else {
	*ppvObj = NULL;
	return ResultFromScode(E_NOINTERFACE);
    }
}

/*
 *----------------------------------------------------------------------
 * TDataObject::AddRef
 *
 *	Increment the reference count.
 *----------------------------------------------------------------------
 */
STDMETHODIMP_(ULONG)
TDataObject::AddRef()
{
    m_refCnt++;
    return m_refCnt;
}

/*
 *----------------------------------------------------------------------
 * TDataObject::Release
 *
 *	Decrement the reference count, destroy if count reaches 0.
 *----------------------------------------------------------------------
 */
STDMETHODIMP_(ULONG)
TDataObject::Release()
{
    m_refCnt--;
    if (m_refCnt == 0L) {
	delete this;
	return 0L;
    }
    return m_refCnt;
}

/*
 *----------------------------------------------------------------------
 * TDataObject::DAdvise
 *
 *	Not implemented and not necessary.
 *----------------------------------------------------------------------
 */
STDMETHODIMP
TDataObject::DAdvise(LPFORMATETC pFormatEtc,
		     DWORD advf,
		     LPADVISESINK pAdvSink,
		     DWORD *pdwConnection)
{
    return ResultFromScode(OLE_E_ADVISENOTSUPPORTED);
}

/*
 *----------------------------------------------------------------------
 * TDataObject::EnumDAdvise
 *
 *	Not implemented and not necessary.
 *----------------------------------------------------------------------
 */
STDMETHODIMP
TDataObject::EnumDAdvise(IEnumSTATDATA **ppEnumAdvise)
{
    return ResultFromScode(OLE_E_ADVISENOTSUPPORTED);
}

/*
 *----------------------------------------------------------------------
 * TDataObject::DUnadvise
 *
 *	Not implemented and not necessary.
 *----------------------------------------------------------------------
 */
STDMETHODIMP
TDataObject::DUnadvise(DWORD dwConnection)
{
    return ResultFromScode(OLE_E_NOCONNECTION);
}

/*
 *----------------------------------------------------------------------
 * TDataObject::GetCanonicalFormatEtc
 *
 *	Not implemented and not necessary.
 *----------------------------------------------------------------------
 */
STDMETHODIMP
TDataObject::GetCanonicalFormatEtc(LPFORMATETC pFormatEtcIn,
				   LPFORMATETC pFormatEtcOut)

{
    return ResultFromScode(E_NOTIMPL);
}

/*
 *----------------------------------------------------------------------
 * TDataObject::SetData
 *
 *	Not implemented and not necessary.
 *----------------------------------------------------------------------
 */
STDMETHODIMP
TDataObject::SetData(LPFORMATETC pFormatEtc, LPSTGMEDIUM pMedium,
		     BOOL fRelease)
{
    return ResultFromScode(E_NOTIMPL);
}

/*
 *----------------------------------------------------------------------
 * TDataObject::EnumFormatEtc
 *
 *	Creates an enumerator.
 *----------------------------------------------------------------------
 */
STDMETHODIMP
TDataObject::EnumFormatEtc(DWORD dwDirection,
			   LPENUMFORMATETC *ppEnumFormatEtc)
{
    if (ppEnumFormatEtc == NULL) {
	return ResultFromScode(E_INVALIDARG);
    }

    if (dwDirection != DATADIR_GET) {
	*ppEnumFormatEtc = NULL;
	return ResultFromScode(E_NOTIMPL);
    }

    *ppEnumFormatEtc = CreateFormatEtcEnumerator(this, m_numTypes, m_typeList);
    if (*ppEnumFormatEtc == NULL) {
	return E_OUTOFMEMORY;
    }

    return S_OK;
};

/*
 *----------------------------------------------------------------------
 * TDataObject::GetData
 *
 *	Gets the data into the storage medium if possible.
 *----------------------------------------------------------------------
 */
STDMETHODIMP
TDataObject::GetData(LPFORMATETC pFormatEtcIn, STGMEDIUM *pMedium)
{
    UINT i;
    char *str;
    int len;
    HGLOBAL hMem;
    char *newStr;
    HRESULT hret;

    for (i = 0; i < m_numTypes; i++) {
	if (m_typeList[i].cfFormat == pFormatEtcIn->cfFormat) {
	    if (pFormatEtcIn->tymed & TYMED_HGLOBAL) {
		break;
	    }
	}
    }
    if (i == m_numTypes) {
	return ResultFromScode(DV_E_FORMATETC);
    }

    hret = m_callbacks->GetData(m_clientData, pFormatEtcIn->cfFormat, &str);
    if (hret != S_OK) {
	return hret;
    }
    len = lstrlen(str) + 1;
    
    hMem = GlobalAlloc(GMEM_SHARE | GMEM_MOVEABLE, len);

    if (hMem == NULL) {
        return ResultFromScode(STG_E_MEDIUMFULL);
    }

    newStr = (char *) GlobalLock(hMem);
    memcpy(newStr, str, len);
    GlobalUnlock(hMem);

    pMedium->hGlobal = hMem;
    pMedium->tymed = TYMED_HGLOBAL;
    pMedium->pUnkForRelease = NULL;
    return S_OK;
}

/*
 *----------------------------------------------------------------------
 * TDataObject::QueryGetData
 *
 *	XXX: Needs to be implemented.
 *----------------------------------------------------------------------
 */
STDMETHODIMP
TDataObject::QueryGetData(LPFORMATETC pFormatEtcIn)
{
    UINT i;

    for (i = 0; i < m_numTypes; i++) {
	if (m_typeList[i].cfFormat == pFormatEtcIn->cfFormat) {
	    if (pFormatEtcIn->tymed & TYMED_HGLOBAL) {
		return S_OK;
	    }
	}
    }

    return ResultFromScode(DV_E_FORMATETC);
}

/*
 *----------------------------------------------------------------------
 * TDataObject::GetDataHere
 *
 *	Not implemented and not necessary.
 *----------------------------------------------------------------------
 */
STDMETHODIMP
TDataObject::GetDataHere(LPFORMATETC pformatetc, LPSTGMEDIUM pmedium)
{
    return ResultFromScode(E_NOTIMPL);
}
