/*
 *----------------------------------------------------------------------
 * Filename: tdropsource.c
 *
 *	Implements the TDropSource class, an IDropSource object
 *
 * Copyright (c) 1995 LAS Incorporated.
 * All rights reserved.
 *
 * Copyright (c) 1995 University of California, Berkeley.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * IN NO EVENT SHALL LAS INCORPORATED BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * LAS INCORPORATED SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND LAS INCORPORATED HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *----------------------------------------------------------------------
 */

#include <windows.h>
#include <ole2.h>
#include <string.h>
#include "tobject.h"

/*
 *----------------------------------------------------------------------
 * TDropSource::TDropSource
 *
 *	Constructor
 *----------------------------------------------------------------------
 */
TDropSource::TDropSource(DWORD keyState, DWORD button)
{
    m_keyState = keyState;
    m_button = button;
    m_refCnt = 0;
    return;
}

/*
 *----------------------------------------------------------------------
 * TDropSource::~TDropSource
 *
 *	Destructor
 *----------------------------------------------------------------------
 */
TDropSource::~TDropSource(void)
{
    return;
}

/*
 *----------------------------------------------------------------------
 * TDropSource::QueryInterface
 *
 *	Does interface negotiation.  IUnknown and IDropSource are
 *	the supported interfaces.
 *
 * Arguments:
 *
 *	REFIID riid	A reference to the interface being queried.
 *      LPVOID *ppvObj	Out parameter to return a pointer to interface.
 *
 * Results:
 *	S_OK if interface is supported, E_NOINTERFACE if it is not
 *
 *----------------------------------------------------------------------
 */
STDMETHODIMP
TDropSource::QueryInterface(REFIID riid, LPVOID *ppv)
{
    if (IsEqualIID(riid, IID_IUnknown) || IsEqualIID(riid, IID_IDropSource))
    {
        *ppv = (LPVOID) this;
	((LPUNKNOWN)*ppv)->AddRef();

	return S_OK;
    }

    *ppv = NULL;
    return ResultFromScode(E_NOINTERFACE);
}

/*
 *----------------------------------------------------------------------
 * TDropSource::AddRef
 *
 *	Increment the reference count on the object as well as on
 *	the unknown object this enumerator is linked to.
 *
 * Results:
 *	The current reference count on the object.
 *----------------------------------------------------------------------
 */
STDMETHODIMP_(ULONG)
TDropSource::AddRef(void)
{
    m_refCnt++;
    return m_refCnt;
}

/*
 *----------------------------------------------------------------------
 * TDropSource::Release
 *
 *	Decrement the reference count on the object as well as on
 *	the unknown object this enumerator is linked to.  If the
 *	reference count becomes 0, the object is deleted.
 *
 * Results:
 *	The current reference count on the object.
 *----------------------------------------------------------------------
 */
STDMETHODIMP_(ULONG)
TDropSource::Release(void)
{
    m_refCnt--;

    if (m_refCnt == 0L) {
        delete this;
	return 0L;
    }

    return m_refCnt;
}

/*
 *----------------------------------------------------------------------
 * TDropSource::QueryDragContinue
 *
 *	Determines whether a drag operation should continue
 *
 * Arguments:
 *	fEscapePressed		Was the escape key pressed? (in)
 *	grfKeyState		Identifies the present state of the modifier
 *				  keys on the keyboard (in)
 *
 * Result:
 *	DRAGDROP_S_CANCEL to stop the drag, DRAGDROP_S_DROP to drop the
 *	data where it is, or S_OK to continue.
 *
 * Notes:
 *	In the current implementation, any change the the modifiers
 *	causes an immediate drop.
 *----------------------------------------------------------------------
 */
STDMETHODIMP
TDropSource::QueryContinueDrag(BOOL fEscapePressed, DWORD keyState)
{
    if (fEscapePressed) {
        return ResultFromScode(DRAGDROP_S_CANCEL);
    }

    if (m_button == 0) {
	if (keyState != m_keyState) {
	    return ResultFromScode(DRAGDROP_S_DROP);
	}
    } else {
	if (! (keyState & m_button)) {
	    return ResultFromScode(DRAGDROP_S_DROP);
	}
    }

    return S_OK;
}

/*
 *----------------------------------------------------------------------
 * TDropSource::GiveFeedback
 *
 *	Enables a source application to provide cursor feedback during a 
 *	drag and drop operation.  
 *
 * Arguments
 *	dwEffect	The DROPEFFECT returned from the most recent call
 *			to IDropTarget::DragEnter or IDropTarget::DragOver
 *
 * Result:
 *	S_OK if this routine sets the cursor, DRAGDROP_S_USEDEFAULTCURSORS
 *	if OLE should use the default cursor to provide feedback.
 *----------------------------------------------------------------------
 */
STDMETHODIMP
TDropSource::GiveFeedback(DWORD dwEffect)
{
    return ResultFromScode(DRAGDROP_S_USEDEFAULTCURSORS);
}
