/*
 *----------------------------------------------------------------------
 * Filename: tdroptarget.cpp
 *
 *	Implements the TDropTarget class, an IDropTarget object
 *
 * Copyright (c) 1995 LAS Incorporated.
 * All rights reserved.
 *
 * Copyright (c) 1995 University of California, Berkeley.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * IN NO EVENT SHALL LAS INCORPORATED BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * LAS INCORPORATED SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND LAS INCORPORATED HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *----------------------------------------------------------------------
 */

#include <windows.h>
#include <ole2.h>
#include <string.h>
#include "tobject.h"

/* #define DEBUG 1 */

/*
 *----------------------------------------------------------------------
 * TDropTarget::TDropTarget
 *
 *	Constructor
 *----------------------------------------------------------------------
 */
TDropTarget::TDropTarget(PTCallback callbacks, LPVOID clientData)
{
    m_callbacks = callbacks;
    m_clientData = clientData;
    m_refCnt = 0;
    return;
}

/*
 *----------------------------------------------------------------------
 * TDropTarget::~TDropTarget
 *
 *	Destructor
 *----------------------------------------------------------------------
 */
TDropTarget::~TDropTarget(void)
{
    return;
}

/*
 *----------------------------------------------------------------------
 * TDropTarget::QueryInterface
 *
 *	Does interface negotiation.  IUnknown and IDropTarget are
 *	the supported interfaces.
 *
 * Arguments:
 *
 *	REFIID riid	A reference to the interface being queried.
 *      LPVOID *ppvObj	Out parameter to return a pointer to interface.
 *
 * Results:
 *	S_OK if interface is supported, E_NOINTERFACE if it is not
 *
 *----------------------------------------------------------------------
 */
STDMETHODIMP
TDropTarget::QueryInterface(REFIID riid, LPVOID *ppv)
{
    if (IsEqualIID(riid, IID_IUnknown) || IsEqualIID(riid, IID_IDropTarget))
    {
        *ppv = (LPVOID) this;
	((LPUNKNOWN)*ppv)->AddRef();

	return S_OK;
    }

    *ppv = NULL;
    return ResultFromScode(E_NOINTERFACE);
}

/*
 *----------------------------------------------------------------------
 * TDropTarget::AddRef
 *
 *	Increment the reference count on the object as well as on
 *	the unknown object this enumerator is linked to.
 *
 * Results:
 *	The current reference count on the object.
 *----------------------------------------------------------------------
 */
STDMETHODIMP_(ULONG)
TDropTarget::AddRef(void)
{
    m_refCnt++;
    return m_refCnt;
}

/*
 *----------------------------------------------------------------------
 * TDropTarget::Release
 *
 *	Decrement the reference count on the object as well as on
 *	the unknown object this enumerator is linked to.  If the
 *	reference count becomes 0, the object is deleted.
 *
 * Results:
 *	The current reference count on the object.
 *----------------------------------------------------------------------
 */
STDMETHODIMP_(ULONG)
TDropTarget::Release(void)
{
    m_refCnt--;

    if (m_refCnt == 0L) {
        delete this;
	return 0L;
    }

    return m_refCnt;
}

#ifdef DEBUG
#include <stdio.h>
static void Log(char *str)
{
    FILE *fp = fopen("c:\\temp\\dnd.log", "a");
    if (fp) {
	fputs(str, fp);
	fclose(fp);
    }
}
#endif
/*
 *----------------------------------------------------------------------
 * TDropTarget::DragEnter
 *
 *	Determines whether the target window can accept the dragged object
 *	and what effect the drop would have on the target window.
 *
 * Arguments:
 *	pIDataObject	An instance of the IDataObject interface through
 *			  which dragged data is accessible. (in)
 *	grfKeyState	Present state of keyboard modifiers (in)
 *	pt		Structure that holds mouse screen coordinates (in)
 *	pdwEffect	Points to where to return the effect of this
 *			  drag operation.
 *
 * Result:
 *	S_OK if everything is OK.
 *
 *----------------------------------------------------------------------
 */
STDMETHODIMP
TDropTarget::DragEnter(LPDATAOBJECT pIDataSource, DWORD grfKeyState,
		       POINTL pt, LPDWORD pdwEffect)
{
    FORMATETC fetc;
    HRESULT hret;

    fetc.ptd = NULL;
    fetc.dwAspect = DVASPECT_CONTENT;
    fetc.lindex = -1;
    fetc.tymed = TYMED_HGLOBAL;

#ifdef DEBUG
    {
	char buf[1024];
	char name[1024];

	Log("---- Drag Enter ----\n");
	for (int i = 0xc000; i < 0xc200; i++) {
	    fetc.cfFormat = i;
	    if (pIDataSource->QueryGetData(&fetc) == S_OK) {
		GetClipboardFormatName(i, name, sizeof(name));
		sprintf(buf, "0x%08x: %s\n", i, name);
		Log(buf);
	    }
	}
    }
#endif

    hret = m_callbacks->Reset(m_clientData);
    hret = m_callbacks->Next(m_clientData, &fetc.cfFormat);
    while (hret == S_OK) {
#ifdef DEBUG
	{
	    char name[1024];
	    char buf[1024];
	    int n;
	    n = GetClipboardFormatName(fetc.cfFormat, name, sizeof(name));
	    sprintf(buf, "Trying 0x%08x: %s\n", fetc.cfFormat, n ? name : "");
	    Log(buf);
	}
#endif
	hret = pIDataSource->QueryGetData(&fetc);
	if (hret == S_OK) {
	    *pdwEffect = m_lastEffect = DROPEFFECT_LINK|DROPEFFECT_COPY;
#ifdef DEBUG
	    Log("   Using format\n");
#endif
	    m_lastFormat = fetc.cfFormat;
	    return Callback(pIDataSource, grfKeyState, pt, pdwEffect,
			    "<DragEnter>");
	}
	hret = m_callbacks->Next(m_clientData, &fetc.cfFormat);
    }

    *pdwEffect = m_lastEffect = DROPEFFECT_NONE;

    return S_OK;
}

/*
 *----------------------------------------------------------------------
 * TDropTarget::DragOver
 *
 *	Provides feedback to the user and to DoDragDrop about the state
 *	of the drag operation within a drop target application.  This
 *	gets called when the mouse has moved within a window that was
 *	already DragEnter'ed.
 *
 * Arguments:
 *	grfKeyState	Present state of the keyboard modifiers
 *	pt		Current mouse screen coordinates.
 *	pdwEffect	Points to where to return the effect of this
 *			  drag operation.
 *
 * Result:
 *	S_OK
 *
 * Notes:
 *	This routine is called on every WM_MOUSEMOVE, so it should be
 *	quick.
 *----------------------------------------------------------------------
 */
STDMETHODIMP
TDropTarget::DragOver(DWORD grfKeyState, POINTL pt, LPDWORD pdwEffect)
{
    FORMATETC fetc;
    HRESULT hret;

    fetc.ptd = NULL;
    fetc.dwAspect = DVASPECT_CONTENT;
    fetc.lindex = -1;
    fetc.tymed = TYMED_HGLOBAL;

    hret = m_callbacks->SetData(m_clientData, m_lastFormat, pt.x, pt.y, NULL,
				"<Drag>");
    *pdwEffect = m_lastEffect;
    return S_OK;
}

/*
 *----------------------------------------------------------------------
 * TDropTarget::DragLeave
 *
 *	This method is called when the mouse leaves the area of a given
 *	target while a drag is in progress or when the drag operation is
 *	canceled.  Causes the drop target to remove its feedback.
 *
 * Arguments:
 *	None
 *
 * Result:
 *	Result from Callback
 *----------------------------------------------------------------------
 */
STDMETHODIMP
TDropTarget::DragLeave(void)
{
    return m_callbacks->SetData(m_clientData, m_lastFormat, 0, 0, NULL,
				"<DragLeave>");
}

/*
 *----------------------------------------------------------------------
 * TDropTarget::Drop
 *
 *	Drops the source data, indicated by pDataObj, on this target
 *	application.
 *
 * Arguments:
 *	pDataObj	Object through which the data is accessible.
 *	grfKeyState	Current keyboard/mouse state.
 *	pt		Mouse screen coordinates
 *	pdwEffect	Action that took place is stored here
 *
 * Result:
 *	Result from Callback
 *----------------------------------------------------------------------
 */
STDMETHODIMP
TDropTarget::Drop(LPDATAOBJECT pIDataSource, DWORD grfKeyState,
		  POINTL pt, LPDWORD pdwEffect)
{
    return Callback(pIDataSource, grfKeyState, pt, pdwEffect, "<Drop>");
}

/*
 *----------------------------------------------------------------------
 * TDropTarget::Callback
 *
 *	Finds the matching event type and makes the callback.  This
 *	is a common utility routine
 *
 * Results:
 *	S_OK if everything is OK.
 *----------------------------------------------------------------------
 */
STDMETHODIMP
TDropTarget::Callback(LPDATAOBJECT pIDataSource, DWORD grfKeyState,
		      POINTL pt, LPDWORD pdwEffect, char *eventStr)
{
    FORMATETC fetc;
    HRESULT hret;
    STGMEDIUM stgMedium;
    char *str;

    fetc.ptd = NULL;
    fetc.dwAspect = DVASPECT_CONTENT;
    fetc.lindex = -1;
    fetc.tymed = TYMED_HGLOBAL;

    hret = m_callbacks->Reset(m_clientData);
    hret = m_callbacks->Next(m_clientData, &fetc.cfFormat);
    while (hret == S_OK) {
	hret = pIDataSource->QueryGetData(&fetc);
	if (hret == S_OK) {
	    break;
	}
	hret = m_callbacks->Next(m_clientData, &fetc.cfFormat);
    }

    if (hret != S_OK) {
	*pdwEffect = DROPEFFECT_NONE;
	return hret;
    }

    stgMedium.tymed = TYMED_HGLOBAL;

    hret = pIDataSource->GetData(&fetc, &stgMedium);
    if (hret == S_OK) {
	str = (char *) GlobalLock(stgMedium.hGlobal);

	hret = m_callbacks->SetData(m_clientData, fetc.cfFormat,
				    pt.x, pt.y, str, eventStr);
	GlobalUnlock(stgMedium.hGlobal);
	if (! stgMedium.pUnkForRelease) {
	    GlobalFree(stgMedium.hGlobal);
	}
    } else {
	*pdwEffect = DROPEFFECT_NONE;
	return hret;
    }

    return S_OK;
}
